/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


#include "quakedef.h"
#include "png.h"

extern int image_width;
extern int image_height;

byte *LoadPNG (FILE *f, char *name)
{
    byte			header[8], **rowpointers, *image_rgba;
    png_structp		png;
    png_infop		pnginfo;
    int				y, width, height, bitdepth, colortype;
    int				interlace, compression, filter, bytesperpixel;
    unsigned long	rowbytes;

    if (!f && COM_FOpenFile(name, &f)) return NULL;

    fread (header, 1, 8, f);

    if (png_sig_cmp(header, 0, 8) != 0)
    {
        Con_Printf ("%s : Invalid PNG file\n", name);
        fclose (f);
        return NULL;
    }

    if (!(png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)))
    {
        fclose (f);
        return NULL;
    }

    if (!(pnginfo = png_create_info_struct(png)))
    {
        png_destroy_read_struct (&png, &pnginfo, NULL);
        fclose (f);
        return NULL;
    }

    if (setjmp(png_jmpbuf(png)))
    {
        png_destroy_read_struct (&png, &pnginfo, NULL);
        fclose (f);
        return NULL;
    }
    png_init_io (png, f);
    png_set_sig_bytes (png, 8);
    png_read_info (png, pnginfo);
    png_get_IHDR (png, pnginfo, (png_uint_32 *)&width, (png_uint_32 *)&height, &bitdepth, &colortype, &interlace, &compression, &filter);

    if (colortype == PNG_COLOR_TYPE_PALETTE)
    {
        png_set_palette_to_rgb (png);
        png_set_filler (png, 255, PNG_FILLER_AFTER);
    }

    if (colortype == PNG_COLOR_TYPE_GRAY && bitdepth < 8)
    {
        png_set_expand_gray_1_2_4_to_8 (png);
    }

    if (png_get_valid(png, pnginfo, PNG_INFO_tRNS))
    {
        png_set_tRNS_to_alpha (png);
    }

    if (colortype == PNG_COLOR_TYPE_GRAY || colortype == PNG_COLOR_TYPE_GRAY_ALPHA)
    {
        png_set_gray_to_rgb (png);
        png_set_filler (png, 255, PNG_FILLER_AFTER);
    }

    if (colortype != PNG_COLOR_TYPE_RGBA)
    {
        png_set_filler (png, 255, PNG_FILLER_AFTER);
    }

    if (bitdepth < 8)
    {
        png_set_expand (png);
    }
    else if (bitdepth == 16)
    {
        png_set_strip_16 (png);
    }
    png_read_update_info (png, pnginfo);
    rowbytes = png_get_rowbytes (png, pnginfo);
    bytesperpixel = png_get_channels (png, pnginfo);
    bitdepth = png_get_bit_depth (png, pnginfo);

    if (bitdepth != 8 || bytesperpixel != 4)
    {
        Con_Printf ("%s : Bad PNG color depth and/or bpp\n", name);
        fclose (f);
        png_destroy_read_struct (&png, &pnginfo, NULL);
        return NULL;
    }
    image_rgba = malloc (height * rowbytes);
    rowpointers = malloc (height * 4);

    for (y=0 ; y<height ; y++)
    {
        rowpointers[y] = image_rgba + y * rowbytes;
    }
    png_read_image (png, rowpointers);
    png_read_end (png, NULL);
    png_destroy_read_struct (&png, &pnginfo, NULL);

    if (rowpointers)
    {
        free (rowpointers);
    }
    fclose (f);
    image_width = width;
    image_height = height;

    return image_rgba;
}

//_____________________________________________________________________________
